/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.objectweb.petals.tools.rmi.common.serializable;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.commons.log.FlowAttributes;

/**
 * The message exchange implementation.
 * @author Nicolas Salatge - EBM WebSourcing
 */
public class MessageExchange implements javax.jbi.messaging.MessageExchange, Cloneable,
        Serializable {

    /**
     * the serialVersionUID.
     */
    private static final long serialVersionUID = -3551387L;

    /**
     * IN_MSG constant.
     */
    public static final String IN_MSG = "in";

    /**
     * IN_ONLY_PATTERN constant.
     */
    public static final URI IN_ONLY_PATTERN = URI.create("http://www.w3.org/2004/08/wsdl/in-only");

    /**
     * IN_OPTIONAL_OUT_PATTERN constant.
     */
    public static final URI IN_OPTIONAL_OUT_PATTERN = URI
            .create("http://www.w3.org/2004/08/wsdl/in-opt-out");

    /**
     * IN_OUT_PATTERN constant.
     */
    public static final URI IN_OUT_PATTERN = URI.create("http://www.w3.org/2004/08/wsdl/in-out");

    /**
     * OUT_MSG constant.
     */
    public static final String OUT_MSG = "out";

    /**
     * ROBUST_IN_ONLY_PATTERN constant.
     */
    public static final URI ROBUST_IN_ONLY_PATTERN = URI
            .create("http://www.w3.org/2004/08/wsdl/robust-in-only");

    /**
     * SERIALIZE_ROLE_CONSUMER constant.
     */
    private static final short SERIALIZE_ROLE_CONSUMER = 0;

    /**
     * SERIALIZE_ROLE_PROVIDER constant.
     */
    private static final short SERIALIZE_ROLE_PROVIDER = 1;

    /**
     * SERIALIZE_STATUS_ACTIVE constant.
     */
    private static final short SERIALIZE_STATUS_ACTIVE = 0;

    /**
     * SERIALIZE_STATUS_DONE constant.
     */
    private static final short SERIALIZE_STATUS_DONE = 1;

    /**
     * SERIALIZE_STATUS_ERROR constant.
     */
    private static final short SERIALIZE_STATUS_ERROR = 2;

    /**
     * the pattern.
     */
    private URI pattern;

    /**
     * the endpoint.
     */
    private ServiceEndpoint endpoint;

    /**
     * the error.
     */
    private Exception error;

    /**
     * the exchangeId.
     */
    private String exchangeId;

    private FlowAttributes flowAttributes;

    /**
     * the fault.
     */
    private Fault fault;

    /**
     * the interface name.
     */
    private QName interfaceName;

    /**
     * the messages.
     */
    private Map<String, NormalizedMessage> messages = new HashMap<String, NormalizedMessage>();

    /**
     * the operation name.
     */
    private QName operation;

    /**
     * the properties.
     */
    private Map<String, Object> properties = new HashMap<String, Object>();

    /**
     * the role.
     */
    private transient Role role;

    /**
     * the service.
     */
    private QName service;

    /**
     * the status.
     */
    private transient ExchangeStatus status;

    /**
     * verify if transaction is ok.
     */
    private boolean transacted;

    /**
     * Create a MessageExchangeImpl This constructor is called by a consumer who
     * wants to initialize an exchange.
     * 
     * @param consumerDeliveryChannel
     */
    public MessageExchange() {

        this.status = ExchangeStatus.ACTIVE;

        this.role = Role.CONSUMER;

    }

    /**
     * @see javax.jbi.messaging.MessageExchange#createFault()
     * @return the fault message
     * @throws MessagingException
     *             Impossible to create the fault
     */
    public Fault createFault() throws MessagingException {
        return new org.objectweb.petals.tools.rmi.common.serializable.Fault();
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#createMessage()
     * @return the normalized message
     * @throws MessagingException
     *             Impossible to create the message
     */
    public NormalizedMessage createMessage() throws MessagingException {
        return new org.objectweb.petals.tools.rmi.common.serializable.NormalizedMessage();
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getEndpoint()
     * @return the service endpoint
     */
    public ServiceEndpoint getEndpoint() {
        return this.endpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getError()
     * @return the exception
     */
    public Exception getError() {
        return this.error;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getExchangeId()
     * @return the exchange id
     */
    public String getExchangeId() {
        return this.exchangeId;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getFault()
     * @return the fault message
     */
    public Fault getFault() {
        return this.fault;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getInterfaceName()
     * @return the interface name
     */
    public QName getInterfaceName() {
        return this.interfaceName;
    }

    /**
     * get the message associated to the reference. It is <b>Not</b> Case
     * Sensitive.
     * 
     * @see javax.jbi.messaging.MessageExchange#getMessage(java.lang.String)
     * @param name
     *            the name of message
     * @return the normalized message
     */
    public NormalizedMessage getMessage(final String name) {
        NormalizedMessage msg = null;
        if (name != null) {
            msg = this.messages.get(name.toLowerCase());
        } else {
            msg = null;
        }
        return msg;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getOperation()
     * @return the opretation
     */
    public QName getOperation() {
        return this.operation;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getPattern()
     * @return the pattern
     */
    public URI getPattern() {
        return this.pattern;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getProperty(java.lang.String)
     * @param name
     *            the property name
     * @return the object property
     */
    public Object getProperty(final String name) {
        return this.properties.get(name);
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getPropertyNames()
     * @return the set of properties
     */
    public Set<?> getPropertyNames() {
        return this.properties.keySet();
    }

    /**
     * (non-Javadoc).
     * 
     * @see javax.jbi.messaging.MessageExchange#getRole()
     * @return the role
     */
    public Role getRole() {
        return this.role;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getService()
     * @return the service.
     */
    public QName getService() {
        return this.service;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getStatus()
     * @return the exchange status
     */
    public javax.jbi.messaging.ExchangeStatus getStatus() {
        return this.status;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#isTransacted()
     * @return true if is transacted
     */
    public boolean isTransacted() {
        return this.transacted;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setEndpoint(java.lang.Exception)
     * @param endpoint
     *            the endpoint
     */
    public void setEndpoint(final ServiceEndpoint endpoint) {
        this.endpoint = endpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setError(java.lang.Exception)
     * @param error
     *            the error
     */
    public void setError(final Exception error) {
        this.error = error;

        this.status = ExchangeStatus.ERROR;

        this.messages.clear();
    }

    /**
     * (non-Javadoc).
     * 
     * @see javax.jbi.messaging.MessageExchange#setFault(javax.jbi.messaging.Fault)
     * @param fault
     *            the fault
     * @throws MessagingException
     *             Impossible to set the fault
     */
    public void setFault(final Fault fault) throws MessagingException {
        this.fault = fault;

        this.messages.clear();
    }

    /**
     * (non-Javadoc).
     * 
     * @see javax.jbi.messaging.MessageExchange#setFault(javax.jbi.messaging.setInterfaceName)
     * @param interfaceName
     *            the interfaceName
     */
    public void setInterfaceName(final QName interfaceName) {
        this.interfaceName = interfaceName;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setMessage(javax.jbi.messaging.NormalizedMessage,
     *      java.lang.String)
     * @param msg
     *            the message
     * @param name
     *            the name
     * @throws MessagingException
     *             impossible to set the message
     */
    public void setMessage(final NormalizedMessage msg, final String name)
            throws MessagingException {
        // no exception was thrown, set the message
        if (msg == null) {
            throw new MessagingException("NormalizedMessage must be non null.");
        }
        if (name == null || name.trim().length() == 0) {
            throw new MessagingException("The message reference must be non null and non empty.");
        }
        this.messages.put(name.toLowerCase(), msg);
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setOperation(javax.xml.namespace.QName)
     * @param name
     *            the name of the operation
     */
    public void setOperation(final QName name) {
        this.operation = name;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setProperty(java.lang.String,
     *      java.lang.Object)
     * @param name
     *            the name
     * @param obj
     *            the property object
     */
    public void setProperty(final String name, final Object obj) {
        this.properties.put(name, obj);
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setService(QName service)
     * @param service
     *            the service name
     */
    public void setService(final QName service) {
        this.service = service;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setStatus(javax.jbi.messaging.ExchangeStatus)
     * @param status
     *            the status
     * @throws MessagingException
     *             Impossible to set the status
     */
    public void setStatus(final ExchangeStatus status) throws MessagingException {
        this.status = status;

        if (this.status != ExchangeStatus.ACTIVE) {
            this.messages.clear();
        }
    }

    /*
     * Added method
     */
    /**
     * Set the ExchangeId.
     * 
     * @param id
     *            the id of message
     * 
     */
    public void setExchangeId(final String id) {
        this.exchangeId = id;
    }

    /**
     * Set the pattern.
     * 
     * @param p
     *            the pattern
     */
    public void setPattern(final URI p) {
        this.pattern = p;
    }

    /**
     * Set the role.
     * 
     * @param r
     *            the role
     */
    public void setRole(final Role r) {
        this.role = r;
    }

    /**
     * Get all messages.
     * 
     * @return all the messages
     */
    public Map<String, NormalizedMessage> getMessages() {
        return this.messages;
    }

    /**
     * @see MessageExchangeImpl#readObject(ObjectInputStream)
     * @param s
     *            the object input stream
     * @throws IOException
     *             impossible to read the object
     */
    protected void readObjectDelegate(final ObjectInputStream s) throws IOException {
        short roleTemp = s.readShort();
        short statusTemp = s.readShort();

        // Role deserialization
        if (roleTemp == 0) {
            this.role = MessageExchange.Role.CONSUMER;
        } else if (roleTemp == 1) {
            this.role = MessageExchange.Role.PROVIDER;
        }

        // Status deserialization
        if (statusTemp == 0) {
            this.status = ExchangeStatus.ACTIVE;
        } else if (statusTemp == 1) {
            this.status = ExchangeStatus.DONE;
        } else if (statusTemp == 2) {
            this.status = ExchangeStatus.ERROR;
        }

        try {
            s.defaultReadObject();
        } catch (ClassNotFoundException e) {
            throw new IOException(e.getClass() + ":" + e.getMessage());
        }
    }

    /**
     * @see MessageExchangeImpl#writeObject(ObjectOutputStream).
     * @param s
     *            the object output stream
     * @throws IOException
     *             impossible to write object delegate
     * 
     */
    protected void writeObjectDelegate(final ObjectOutputStream s) throws IOException {
        
        // Role serialization
        if (MessageExchange.Role.CONSUMER.equals(this.role)) {
            s.writeShort(MessageExchange.SERIALIZE_ROLE_CONSUMER);
        } else {
            s.writeShort(MessageExchange.SERIALIZE_ROLE_PROVIDER);
        }
        
        // Status serialization
        if (ExchangeStatus.ACTIVE.equals(this.status)) {
            s.writeShort(MessageExchange.SERIALIZE_STATUS_ACTIVE);
        } else if (ExchangeStatus.DONE.equals(this.status)) {
            s.writeShort(MessageExchange.SERIALIZE_STATUS_DONE);
        } else {
            s.writeShort(MessageExchange.SERIALIZE_STATUS_ERROR);
        }
        s.defaultWriteObject();
    }

    /**
     * Used to deserialize Status and Role objects.
     * 
     * @param s
     *            the object input stream
     * @throws IOException
     *             impossible to read object
     */
    private void readObject(final ObjectInputStream s) throws IOException {
        this.readObjectDelegate(s);
    }

    /**
     * Used to serialize Status and Role objects.
     * 
     * @param s
     *            the object output stream
     * @throws IOException
     *             impossible to write object
     */
    private void writeObject(final ObjectOutputStream s) throws IOException {
        this.writeObjectDelegate(s);
    }

    public FlowAttributes getFlowAttributes() {
        return this.flowAttributes;
    }

    public void setFlowAttributes(FlowAttributes flowAttributes) {
        this.flowAttributes = flowAttributes;
    }
}
