/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.objectweb.petals.tools.rmi.common.serializable;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.SocketOptions;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.activation.DataHandler;
import javax.mail.util.ByteArrayDataSource;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import com.ebmwebsourcing.easycommons.pooling.PoolException;
import com.ebmwebsourcing.easycommons.xml.Transformers;

/**
 * @author Adrien Louis - EBM WebSourcing
 */
public class MessageExchangeSerializer {


    private MessageExchangeSerializer() {
        super();
    }

    protected static final MessageExchangeSerializer instance = new MessageExchangeSerializer();

    public static MessageExchangeSerializer instance() {
        return instance;
    }

    public void serializeContent(Source content, ObjectOutputStream s, boolean compress) 
        throws IOException {
        synchronized (content) {
            /*
             * Compress content if specified
             */
            StreamResult sresult = null;
            GZIPOutputStream gz = null;
            if (compress) {
                gz = new GZIPOutputStream(s);
                sresult = new StreamResult(gz);

            } else {
                sresult = new StreamResult(s);
            }

            Transformer transformer = null;
            try {
                transformer = Transformers.takeTransformer();
                transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
                transformer.setOutputProperty(OutputKeys.METHOD, "xml");
                transformer.transform(content, sresult);
            } catch (TransformerException e) {
                throw new IOException(e);
            } catch (PoolException e) {
                throw new IOException(e);
            } finally {
                if (transformer != null) {
                    Transformers.releaseTransformer(transformer);
                }
            }

            if (compress) {
                gz.finish();
            }
        }
    }

    public Source deserializeContent(ObjectInputStream s, boolean compress) throws IOException {
        StreamSource source = new StreamSource();

        InputStream in = s;
        if (compress) {
            in = new GZIPInputStream(s);
        }

        byte[] buff = new byte[1024];
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        int n = 0;
        while (-1 != (n = in.read(buff))) {
            outputStream.write(buff, 0, n);
        }

        ByteArrayInputStream result = new ByteArrayInputStream(outputStream.toByteArray());

        source.setInputStream(result);
        return source;
    }

    /**
     * Serialize all datahandler in the Map. <br>
     * Serialized information is : <br>
     * -attachmentName <br>
     * -attachmentContentType <br>
     * -name <br>
     * -inputStream available bytes count <br>
     * -inputStream as a byteArray
     * 
     * @throws IOException 
     */
    public void serializeAttachments(Map<String, DataHandler> attachments, ObjectOutputStream s) throws IOException {
        s.writeInt(attachments.size());

        for (Entry<String, DataHandler> attachment : attachments.entrySet()) {

            DataHandler value = attachment.getValue();
            if(value != null) {
                InputStream binary = value.getInputStream();
                if (binary != null) {
                    synchronized (binary) {
    
                        int available = binary.available();
    
                        // write information about the attachment
                        s.writeUTF(attachment.getKey());
                        s.writeUTF(value.getContentType());
                        s.writeUTF(value.getName());
                        s.writeInt(available);
    
                        // write attachment content with the default socket output
                        // buffer size
                        byte[] buff = new byte[SocketOptions.SO_SNDBUF];
                        int n = 0;
                        while (-1 != (n = binary.read(buff))) {
                            s.write(buff, 0, n);
                            s.flush();
                        }
                    }
                }
            }
        }
    }

    /**
     * Deserialize all datahandlers in the stream and return a Map. <br>
     * Deserialized information is : <br>
     * -attachmentName <br>
     * -attachmentContentType <br>
     * -name <br>
     * -Stream available bytes count <br>
     * -Stream as a byteArray
     * 
     * @throws IOException
     */
    public Map<String, DataHandler> deserializeAttachments(ObjectInputStream s) throws IOException {
        HashMap<String, DataHandler> attachments = new HashMap<String, DataHandler>();

        int attachmentsCount = s.readInt();

        for (int i = 0; i < attachmentsCount; i++) {

            // read information
            String attachmentName = s.readUTF();
            String contentType = s.readUTF();
            String name = s.readUTF();
            int available = s.readInt();

            // read content
            byte[] content = new byte[available];
            s.readFully(content);
            ByteArrayDataSource dataSource = new ByteArrayDataSource(content, contentType);
            dataSource.setName(name);

            // create the dataHandler
            DataHandler dataHandler = new DataHandler(dataSource);

            attachments.put(attachmentName, dataHandler);
        }
        return attachments;
    }

}
